CON
  ' Set up the processor clock in the standard way for 80MHz
  _CLKMODE = xtal1 + pll8x
  _XINFREQ = 10_000_000 + 0000

PUB start
  cognew(@entry,0)
  repeat

DAT                     org
entry                   jmp     #initialization         'Jump past the constants

' I want to explain some important constants up front, so I'll declare them here.
' Ordinarily they'd be tucked away after the code.  There is also a lot of skipping
' arounf between CON and DAT sections.  I'm declaring them in the order I can best explain them.
'
' I've used some naming conventions.  These prefixes mean:
' NTSC_ :   The value is a constant of the NTSC system. They would probably be the same no matter
'           what resolution or color mode you are using.  You'd probably only change them if you were
'           converting to PAL.  In which case, good luck to you.
' CHOOSE_ : A choice was made for this particular resolution and color depth.  You might change these.
' CALC_     These were calculated from one or more of the choices you made.  Choose again and you may
'           have to change these.

'***************************************************
'* Color Frequency                                 *
'***************************************************
'
' The NTSC output is a signal with various elements that need timing relative to each other.  So we need some
' sort of clock from which to time all the elements.  The frequency of the color carrier is useful.  This is
' 3.579545 Mhz. A cycle of this clock will take 1sec/3.579545MHz = 279.365ns.

CON  NTSC_color_frequency       =     3_579_545
DAT  NTSC_color_freq            long  NTSC_color_frequency

' Additionally this period of 279ns is divided up into 16 by a Phased Locked Loop circuit (PLL), and it is multiples
' this period that the Video Scale Hardware Register (VSCL) is programmed.  This period of 17.460313ns is called a
' "clock".



'***************************************************
'* Horizontal sync                                 *
'***************************************************
'* A hsync takes 10.9us.  That's 10.9us/17.460313ns = 624 clocks (rounded to an integer).
CON  NTSC_hsync_clocks          =               624

' We are going to send each horizontal sync with a single WAITVID.  We'll aproximate the waveform with
' 16 x 4 "color" pixels.  So we need to program the Video Scale register VSCL.  That looks like this:
'
' %xxxx_xxxx_xxxx_________________________ : Not used
' %_______________PPPP_PPPP_______________ : Clocks per pixel (CPP)
' %_________________________FFFF_FFFF_FFFF : Clocks per frame (CPF)
'
' The number of pixels we get is CPF/CPP.  So for this usage CPS is 624/16 = 39.
DAT  NTSC_hsync_VSCL            long  NTSC_hsync_clocks >> 4 << 12 + NTSC_hsync_clocks

' The hsync is made up of 3 different sorts of signal.  We put them into a palette:
' Color0 = Color NTSC_control_signal_palette (yellow hue at zero value) = $8a
' Color1 = Black = $02      
' Color2 = Blanking level = $00
' Color4 = don't care
' Note that the palette is programmed in reverse order:
DAT  NTSC_control_signal_palette long  $00_00_02_8a

' Then we construct a hsync out of 16 pixels of those control signals.
' Note that this is shifted out of the VSU to the right, so LSB goes first.
DAT  NTSC_hsync_pixels          long  %%11_0000_1_2222222_11



'***************************************************
'* Blank lines                                     *
'***************************************************
' The rest of the line after a hsync takes 52.6us.  That's 3008 clocks (to the nearest multiple of 16).
CON  NTSC_active_video_clocks   =     3008

' For blank lines, were' going to programme the VSCL so it slowly outputs 16 pixels over the entire
' length of the active line.  It's all going to be one colour so we don't even need to programme the
' clocks per pixel part, just the clocks per frame.
DAT  NTSC_active_video_VSCL     long  NTSC_active_video_clocks



'***************************************************
'* User graphics lines                             *
'***************************************************
' The important lines at last.  To fit 256 pixels in, we're going to make them 10 clocks each.
' That's 2560 clocks for the user graphics width.  You could use 11 clocks each or some other near value.
' It depends whether you want overscan to the left and right.  
CON CHOOSE_clocks_per_gfx_pixel = 10

' Because we're going to use 256 color mode, we're only going to output 4 pixels per WAITVID.  So that
' decides the number of clocks per frame.  4.
CON CALC_clocks_per_gfx_frame   =     CHOOSE_clocks_per_gfx_pixel*4

' Program the VSCL as before.
DAT CALC_user_data_VSCL         long  CHOOSE_clocks_per_gfx_pixel << 12 + CALC_clocks_per_gfx_frame

' So if we're doing 256 pixels, 4 at a time, that's 256/4 = 64 frames.  64 WAITVIDS.
CON CALC_frames_per_gfx_line    = 256/4

' There is some extra active video that is not used for user graphics.  This is the overscan.
CON CALC_overscan = NTSC_active_video_clocks-CALC_frames_per_gfx_line*CALC_clocks_per_gfx_frame

' It maybe that your TV doesn't centre the picture.  We can tweak the overscan left and right to make
' it balance.  28 works quite well on my TV.  It may well be different for yours.
CON CHOOSE_horizontal_offset    = 28

' So we can work out how much overscan to do left (backporch) and right (frontporch).
CON CALC_backporch = CALC_overscan/2+28
CON CALC_frontporch = CALC_overscan - CALC_overscan/2-28

DAT

'***************************************************
'* The code                                        *
'***************************************************
' Start of real code

initialization          'set up VCFG

                        ' VCFG: setup Video Configuration register and 3-bit tv DAC pins to output
                        movs    VCFG, #%0000_0111       ' VCFG'S = pinmask (pin31: 0000_0111 : pin24)
                        movd    VCFG, #3                ' VCFG'D = pingroup (grp. 3 i.e. pins 24-31)

                        movi    VCFG, #%0_10_111_000    ' baseband video on bottom nibble, 2-bit color, enable chroma on broadcast & baseband
                                                        ' %0_xx_x_x_x_xxx : Not used
                                                        ' %x_10_x_x_x_xxx : Composite video to top nibble, broadcast to bottom nibble
                                                        ' %x_xx_1_x_x_xxx : 4 color mode
                                                        ' %x_xx_x_1_x_xxx : Enable chroma on broadcast
                                                        ' %x_xx_x_x_1_xxx : Enable chroma on baseband
                                                        ' %x_xx_x_x_x_000 : Broadcast Aural FM bits (don't care)

                        or      DIRA, tvport_mask       ' set DAC pins to output

                        ' 
                        or      DIRA, #1                ' enable debug LED
                        mov     OUTA, #1                ' light up debug LED

                        ' CTRA: setup Frequency to Drive Video
                        movi    CTRA,#%00001_111        ' pll internal routed to Video, PHSx+=FRQx (mode 1) + pll(16x)
                        mov     r1, NTSC_color_freq     ' r1: Color frequency in Hz (3.579_545MHz)
                        rdlong  v_clkfreq, #0           ' copy clk frequency. (80Mhz)
                        mov     r2, v_clkfreq           ' r2: CLKFREQ (80MHz)
                        call    #dividefract            ' perform r3 = 2^32 * r1 / r2
                        mov     v_freq, r3              ' v_freq now contains frqa.       (191)
                        mov     FRQA, r3                ' set frequency for counter

'-----------------------------------------------------------------------------
                        'NTSC has 244 "visible" lines, but some of that is overscan etc.
                        '  so pick a number of lines for user graphics e.g. 192
                        '  then display (244-192)/2 = 26 lines of overscan before
                        '  and after the user display.
frame_loop
                        mov     line_loop, #26          '(26 so far)
                        'There's always a horizontal sync at the start of each line.  The clever stuff is
                        '  in the constants described earlier.
:vert_back_porch        mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, hsync
                        'This is just a blank line.  I've made it brown just to show it's there
                        '  but you'd probably want it black.  Notice how the whole line (apart from the
                        '  hsyc is output with one WAITVID.
                        mov     VSCL, NTSC_active_video_VSCL
                        waitvid brown_border_in_color0, #0
                        djnz    line_loop, #:vert_back_porch
'-----------------------------------------------------------------------------
                        'Time to do the user graphics lines.  We're having 192 of them.
                        mov     line_loop, #192         '(218 so far)
                        'hsync                        
user_graphics_lines     mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, hsync
                        'The overscan before the user graphics.  Green in this case. 
                        mov     VSCL,#CALC_backporch
                        waitvid green_border_in_color0, #0 

                        'And now at long last, the user graphics.  Program the VSCL for
                        '  4 pixels per frame.  
                        mov     VSCL, CALC_user_data_VSCL
                        
                        'Output for color bar lines.  Let's make it really dumb and obvious where
                        '  all the work is done by showing all 64 frames for a single line.
                        '  The thing to reall take notice of here is how we do 256 color mode.
                        '  We've programmed for 4 pixels per WAITVID, and do we always make those
                        '  colors 3,2,1 and 0.  Then whatever we put in the palette (the first
                        '  parameter of WAITVID is what we get.  The numbers are backwards to
                        '  cancel out the backwardness of VSU serialisation.  It means that
                        '  you write pixels in the palette in the same order you want them on
                        '  screen.
                        cmp     line_loop,#20  wc
              if_a      jmp     #top_flag
                        waitvid colorbar0,#%%3210
                        waitvid colorbar1,#%%3210
                        waitvid colorbar2,#%%3210
                        waitvid colorbar3,#%%3210
                        waitvid colorbar4,#%%3210
                        waitvid colorbar5,#%%3210
                        waitvid colorbar6,#%%3210
                        waitvid colorbar7,#%%3210
                        waitvid colorbar8,#%%3210
                        waitvid colorbar9,#%%3210
                        waitvid colorbar10,#%%3210
                        waitvid colorbar11,#%%3210
                        waitvid colorbar12,#%%3210
                        waitvid colorbar13,#%%3210
                        waitvid colorbar14,#%%3210
                        waitvid colorbar15,#%%3210
                        waitvid colorbar16,#%%3210
                        waitvid colorbar17,#%%3210
                        waitvid colorbar18,#%%3210
                        waitvid colorbar19,#%%3210
                        waitvid colorbar20,#%%3210
                        waitvid colorbar21,#%%3210
                        waitvid colorbar22,#%%3210
                        waitvid colorbar23,#%%3210
                        waitvid colorbar24,#%%3210
                        waitvid colorbar25,#%%3210
                        waitvid colorbar26,#%%3210
                        waitvid colorbar27,#%%3210
                        waitvid colorbar28,#%%3210
                        waitvid colorbar29,#%%3210
                        waitvid colorbar30,#%%3210
                        waitvid colorbar31,#%%3210
                        waitvid colorbar32,#%%3210
                        waitvid colorbar33,#%%3210
                        waitvid colorbar34,#%%3210
                        waitvid colorbar35,#%%3210
                        waitvid colorbar36,#%%3210
                        waitvid colorbar37,#%%3210
                        waitvid colorbar38,#%%3210
                        waitvid colorbar39,#%%3210
                        waitvid colorbar40,#%%3210
                        waitvid colorbar41,#%%3210
                        waitvid colorbar42,#%%3210
                        waitvid colorbar43,#%%3210
                        waitvid colorbar44,#%%3210
                        waitvid colorbar45,#%%3210
                        waitvid colorbar46,#%%3210
                        waitvid colorbar47,#%%3210
                        waitvid colorbar48,#%%3210
                        waitvid colorbar49,#%%3210
                        waitvid colorbar50,#%%3210
                        waitvid colorbar51,#%%3210
                        waitvid colorbar52,#%%3210
                        waitvid colorbar53,#%%3210
                        waitvid colorbar54,#%%3210
                        waitvid colorbar55,#%%3210
                        waitvid colorbar56,#%%3210
                        waitvid colorbar57,#%%3210
                        waitvid colorbar58,#%%3210
                        waitvid colorbar59,#%%3210
                        waitvid colorbar60,#%%3210
                        waitvid colorbar61,#%%3210
                        waitvid colorbar62,#%%3210
                        waitvid colorbar63,#%%3210
                        jmp     #end_of_flag_line          

                        'Code for the lines that are part of the flag.  
top_flag                cmp    line_loop,#75  wc
              if_a      jmp     #middle_flag

                        mov     tile_loop,#24
:flag1                  waitvid user_palette3,#%%3210
                        djnz    tile_loop,#:flag1

                        mov tile_loop,#16
:flag2                  waitvid user_palette2,#%%3210
                        djnz    tile_loop,#:flag2
                        
                        mov tile_loop,#24
:flag3                  waitvid user_palette3,#%%3210
                        djnz    tile_loop,#:flag3
                        jmp     #end_of_flag_line

middle_flag             cmp     line_loop,#117  wc
              if_a      jmp     #bottom_flag

                        mov    tile_loop,#64
middle_flag_loop        waitvid user_palette2,#%%3210
                        djnz    tile_loop,#middle_flag_loop
                        jmp    #end_of_flag_line
bottom_flag
                        mov tile_loop,#24
:flag1                  waitvid user_palette3,#%%3210
                        djnz    tile_loop,#:flag1

                        mov tile_loop,#16
:flag2                  waitvid user_palette2,#%%3210
                        djnz    tile_loop,#:flag2
                        
                        mov tile_loop,#24
:flag3                  waitvid user_palette3,#%%3210
                        djnz    tile_loop,#:flag3

                        'We've finished drawing a user graphics line, whether it was color
                        '  bar or flag.  Now do the overscan bit.  This time in blue.
end_of_flag_line        mov    VSCL,#CALC_frontporch
                        waitvid blue_border_in_color0 , border 
                        'loop
                        djnz    line_loop, #user_graphics_lines
'-----------------------------------------------------------------------------
                        'Overscan at the bottom of screen.  Same as top of screen, but
                        '  this time in magenta.
                        mov     line_loop, #26          '(244)
                        'hsync
vert_front_porch        mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, hsync
                        mov     VSCL, NTSC_active_video_VSCL
                        waitvid magenta_border_in_color0, #0
                        djnz    line_loop, #vert_front_porch
'-----------------------------------------------------------------------------
                        'This is the vertical sync.  It consists of 3 sections of 6 lines each.
                        mov     line_loop, #6           '(250)
:vsync_higha            mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, vsync_high_1
                        mov     VSCL, NTSC_active_video_VSCL
                        waitvid NTSC_control_signal_palette, vsync_high_2
                        djnz    line_loop, #:vsync_higha
'-----------------------------------------------------------------------------
                        mov     line_loop, #6           '(256)
:vsync_low              mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, vsync_low_1
                        mov     VSCL, NTSC_active_video_VSCL
                        waitvid NTSC_control_signal_palette, vsync_low_2
                        djnz    line_loop, #:vsync_low
'-----------------------------------------------------------------------------
                        mov     line_loop, #6           '(250)
:vsync_highb            mov     VSCL, NTSC_hsync_VSCL
                        waitvid NTSC_control_signal_palette, vsync_high_1
                        mov     VSCL, NTSC_active_video_VSCL
                        waitvid NTSC_control_signal_palette, vsync_high_2
                        djnz    line_loop, #:vsync_highb
'-----------------------------------------------------------------------------
                        mov     OUTA, #0

                        jmp     #frame_loop

' General Purpose Registers
r0                      long                    $0                              ' should typically equal 0
r1                      long                    $0
r2                      long                    $0
r3                      long                    $0

' Video (TV) Registers
tvport_mask             long                    %0000_0111<<24

v_freq                  long                    0

' Graphics related vars.
v_coffset               long                    $02020202                       ' color offset (every color is added by $02)
v_clkfreq               long                    $0

' /////////////////////////////////////////////////////////////////////////////
' dividefract:
' Perform 2^32 * r1/r2, result stored in r3 (useful for TV calc)
' This is taken from the tv driver.
' NOTE: It divides a bottom heavy fraction e.g. 1/2 and gives the result as a 32-bit fraction.
' /////////////////////////////////////////////////////////////////////////////
dividefract                                     
                        mov     r0,#32+1
:loop                   cmpsub  r1,r2           wc
                        rcl     r3,#1
                        shl     r1,#1
                        djnz    r0,#:loop

dividefract_ret         ret                             '+140


'Pixel streams
'  These are shifted out of the VSU to the right, so lowest bits are actioned first.
'
hsync                   long    %%11_0000_1_2222222_11  ' Used with NTSC_control_signal_palette so:
                                                        '      0 = blanking level
                                                        '      1 = Black
                                                        '      2 = Color NTSC_control_signal_palette (yellow at zero value)
vsync_high_1            long    %%11111111111_222_11
vsync_high_2            long    %%1111111111111111
vsync_low_1             long    %%22222222222222_11
vsync_low_2             long    %%1_222222222222222
all_black               long    %%1111111111111111
border                  long    %%0000000000000000

' Some unimportant irrelevant constants for generating demo user display etc.
line_loop               long    0
tile_loop               long    0
stripe1                 long    %%1111111100000000
stripe2                 long    %%0000000011111111

'Palettes
'  These are always 4 colors (or blanking level) stored in reverse order:
'                               Color3_Color2_Color1_Color0
'
blue_border_in_color0   long    $00_00_00_da
green_border_in_color0  long    $00_00_00_aa
magenta_border_in_color0 long   $00_00_00_3a
brown_border_in_color0  long    $00_00_00_7a
user_palette2           long    $5a_5a_5a_5a
user_palette3           long    $05_05_05_05
colorbar0               long    $0b_0b_0b_0b
colorbar1               long    $0c_0c_0c_0c
colorbar2               long    $0d_0d_0d_0d
colorbar3               long    $0e_0e_0e_0e
colorbar4               long    $1b_0b_1b_1b
colorbar5               long    $1c_0c_1c_1c
colorbar6               long    $1d_0d_1d_1d
colorbar7               long    $1e_0e_1e_1e
colorbar8               long    $2b_0b_2b_2b
colorbar9               long    $2c_2c_2c_2c
colorbar10              long    $2d_2d_2d_2d
colorbar11              long    $2e_2e_2e_2e
colorbar12              long    $3b_3b_3b_3b
colorbar13              long    $3c_3c_3c_3c
colorbar14              long    $3d_3d_3d_3d
colorbar15              long    $3e_3e_3e_3e
colorbar16              long    $4b_4b_4b_4b
colorbar17              long    $4c_4c_4c_4c
colorbar18              long    $4d_4d_4d_4d
colorbar19              long    $4e_4e_4e_4e
colorbar20              long    $5b_5b_5b_5b
colorbar21              long    $5c_5c_5c_5c
colorbar22              long    $5d_5d_5d_5d
colorbar23              long    $5e_5e_5e_5e
colorbar24              long    $6b_6b_6b_6b
colorbar25              long    $6c_6c_6c_6c
colorbar26              long    $6d_6d_6d_6d
colorbar27              long    $6e_6e_6e_6e
colorbar28              long    $7b_7b_7b_7b
colorbar29              long    $7c_7c_7c_7c
colorbar30              long    $7d_7d_7d_7d
colorbar31              long    $7e_7e_7e_7e
colorbar32              long    $8b_8b_8b_8b
colorbar33              long    $8c_8c_8c_8c
colorbar34              long    $8d_8d_8d_8d
colorbar35              long    $8e_8e_8e_8e
colorbar36              long    $9b_9b_9b_9b
colorbar37              long    $9c_9c_9c_9c
colorbar38              long    $9d_9d_9d_9d
colorbar39              long    $9e_9e_9e_9e
colorbar40              long    $ab_ab_ab_ab
colorbar41              long    $ac_ac_ac_ac
colorbar42              long    $ad_ad_ad_ad
colorbar43              long    $ae_ae_ae_ae
colorbar44              long    $bb_bb_bb_bb
colorbar45              long    $bc_bc_bc_bc
colorbar46              long    $bd_bd_bd_bd
colorbar47              long    $be_be_be_be
colorbar48              long    $cb_cb_cb_cb
colorbar49              long    $cc_cc_cc_cc
colorbar50              long    $cd_cd_cd_cd
colorbar51              long    $ce_ce_ce_ce
colorbar52              long    $db_db_db_db
colorbar53              long    $dc_dc_dc_dc
colorbar54              long    $dd_dd_dd_dd
colorbar55              long    $de_de_de_de
colorbar56              long    $eb_eb_eb_eb
colorbar57              long    $ec_ec_ec_ec
colorbar58              long    $ed_ed_ed_ed
colorbar59              long    $ee_ee_ee_ee
colorbar60              long    $fb_fb_fb_fb
colorbar61              long    $fc_fc_fc_fc
colorbar62              long    $fd_fd_fd_fd
colorbar63              long    $fe_fe_fe_fe

                                   